#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <bfd.h>

#ifdef WIN32
# include <windows.h>
# include <imagehlp.h>
# include <tlhelp32.h>
//# include "dbghelp.h"
#else
# include <dlfcn.h>
// this cond is from ruby
# if defined(__hp9000s300) ||  (defined(__NetBSD__) && !defined(__ELF__)) || defined(__BORLANDC__) || (defined(__FreeBSD__) && !defined(__ELF__)) || (defined(__OpenBSD__) && !defined(__ELF__)) || defined(NeXT) || defined(__WATCOMC__) || defined(__APPLE__)
# include <mach-o/dyld.h>
#else
# include <link.h>
#endif
#endif

#ifdef WIN32
static void winerror(const char* msg) {
    LPVOID lpMsgBuf;

    //˥å

    FormatMessage(              //顼ɽʸ
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL, GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf, 0, NULL);

    printf("%s: %s", msg, (const char*)lpMsgBuf);
//    MessageBox(NULL, (const char*)lpMsgBuf, NULL, MB_OK);   //åɽ

    LocalFree(lpMsgBuf);
}
#endif

typedef void (*regsym_cb) (const char*, void*);

static void add_symbol(regsym_cb fp, int offset, asymbol* asym) {
    const char* name;
    void* value;

    name = bfd_asymbol_name(asym);
    value = (void*)bfd_asymbol_value(asym);

    if (name == NULL || strlen(name) == 0) return;

    if (value == 0) return;

#if 0
#ifdef WIN32
    if (name[0] != '_') return;
    name++;
/*
    if (offset) {
        HMODULE cur_mod = (HMODULE)offset;
        offset = 0;
        value = GetProcAddress(cur_mod, name);
    }
*/
#endif

#ifdef __APPLE__
    if (name[0] != '_') return;
    name++;
    if (strchr(name, ':')) return;
#endif
#endif

//    if (!isalpha((int)name[0])) return;

    if ((unsigned int)value < (unsigned int)offset) {
        value += offset;
    }
//    printf("[C] %x %s\n", value, name);
    fp(name, value);
}

#ifndef WIN32
static int read_symbols(regsym_cb fp, const char* filename, int offset) {
    bfd* abfd;
    int i;
    int symnum;

//    printf("reading %s(0x%x)...\n", filename, offset);

    abfd = bfd_openr(filename, NULL);
    if (abfd == NULL) {
        bfd_perror("bfd_openr");
        return 1;
    }
    if (!bfd_check_format(abfd, bfd_object)) {
        bfd_perror("bfd_check_format");
        return 1;
    }

    {
        void* minisyms;
        unsigned int size;
        char* p;
        asymbol* store;
        int dyn = 0;

        store = bfd_make_empty_symbol(abfd);

        symnum = bfd_read_minisymbols(abfd, dyn, &minisyms, &size);
#if defined(__linux__) || defined(__FreeBSD__)
        if (symnum == 0) {
            dyn = 1;
            symnum = bfd_read_minisymbols(abfd, dyn, &minisyms, &size);
        }
#endif
        if (symnum < 0) {
            bfd_perror("bfd_read_minisymbols");
            goto error;
        }

        p = (char*)minisyms;
        for (i = 0; i < symnum; i++) {
            asymbol* sym;
            sym = bfd_minisymbol_to_symbol(abfd, dyn, p, store);
            add_symbol(fp, offset, sym);
            p += size;
        }

/*
        printf("load %d/%d symbols. (total %d)\n",
               ref->sym_num-prev_symnum, symnum, ref->sym_num);
*/

        free(minisyms);
    }

    bfd_close(abfd);
    return 0;

error:
    bfd_close(abfd);
    return 1;
}
#endif

#if defined(WIN32)
/*
static int load_arch(regsym_cb fp, MODULEENTRY32* me) {
    char buf[1024];
    FILE* f;
    if (!GetModuleFileName(me->hModule, buf, 1024)) {
        winerror("GetModuleFileName");
        return 1;
    }
    f = fopen(buf, "rb");
    if (f) {
        fclose(f);
        return read_symbols(fp, buf, (int)me->modBaseAddr);
    }
    printf("%s: not found. Is it WINE?\n", buf);
    return 0;
}
*/
static int regist_loaded(LPSTR name, ULONG addr, ULONG size, PVOID ctx) {
    regsym_cb fp = (regsym_cb*)ctx;
//    printf("[C] %x %s\n", addr, name);
    fp(name, (void*)addr);
    return TRUE;
}

static int init_arch(regsym_cb fp) {
    HANDLE h;
    DWORD base;
    HANDLE ss;
    MODULEENTRY32 me;
    char buf[4096];

    h = GetCurrentProcess();
    if (!SymInitialize(h, NULL, FALSE)) {
        winerror("SymInitialize");
        return 1;
    }
    ss = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, 0);
    if (!ss) {
        perror("CreateToolhelp32Snapshot");
        return 1;
    }
    me.dwSize = 1024;
    if (!Module32First(ss, &me)) {
        perror("Module32First");
        return 1;
    }
    GetModuleFileName(me.hModule, buf, 4096);
    base = SymLoadModule(h, NULL, buf, NULL, 0, 0);
    if (base == 0) {
        winerror("SymLoadModule");
        return 1;
    }
    if (!SymEnumerateSymbols(h, base, regist_loaded, fp)) {
        winerror("SymEnemerateSymbols");
        return 1;
    }
/*
    if (!SymUnloadModule(h, base)) {
        winerror("SymUnloadModule");
        return 1;
    }
*/

//    SymEnemerateSymbols(h, 
/*
    while (1) {
        if (!SymGetSymNext(h, &sym)) {
            winerror("SymGetSymNext");
            return 1;
        }
        fp(sym.Name, sym.Address);
    }
*/

/*
//    printf("entering win32 initialize\n");
    ss = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, 0);
    if (!ss) {
        perror("CreateToolhelp32Snapshot");
        return 1;
    }
    me.dwSize = 1024;
    if (!Module32First(ss, &me)) {
        perror("Module32First");
        return 1;
    }
    if (load_arch(fp, &me)) return 1;
    for (;;) {
        me.dwSize = 1024;
        if (!Module32Next(ss, &me)) {
            return 0;
        }
        if (load_arch(fp, &me)) return 1;
    }
*/
    return 0;
}
#elif defined(__linux__)
static int regist_loaded(struct dl_phdr_info* info, size_t size, void* data) {
    if (!info->dlpi_name || !info->dlpi_name[0]) return 0;
    return read_symbols(data, info->dlpi_name, info->dlpi_addr);
}

static int init_arch(regsym_cb fp) {
    FILE* f;
    if ((f = fopen("/proc/self/exe", "rb"))) {
        fclose(f);
        if (read_symbols(fp, "/proc/self/exe", 0)) return 1;
    }
    return dl_iterate_phdr(regist_loaded, fp);
}
#elif defined(__FreeBSD__) || (__sun__)
static int init_arch(regsym_cb fp) {
    struct link_map* lmap;
    if (dlinfo(RTLD_SELF, RTLD_DI_LINKMAP, &lmap) == 0) {
        while (lmap) {
            if (read_symbols(fp, lmap->l_name, lmap->l_addr)) return 1;
            lmap = lmap->l_next;
        }
        return 0;
    }
    else {
        perror("dlinfo");
        return 1;
    }
}
#elif defined(__APPLE__)
static int init_arch(regsym_cb fp) {
    int i;
    unsigned long cnt;

    cnt = _dyld_image_count();

    for (i = 0; i < cnt; i++) {
        const char* fname = _dyld_get_image_name(i);
        unsigned long vm_start = _dyld_get_image_vmaddr_slide(i);
        if (read_symbols(fp, fname, vm_start)) return 1;
    }

    return 0;
}
#else
#endif

int reflection_init(regsym_cb fp, const char* exe) {
    int ret;

#ifndef WIN32
    bfd_init();
#endif

#if defined(WIN32) || defined(__linux__) || defined(__FreeBSD__) || defined(__APPLE__) || defined(__sun__)
    ret = init_arch(fp);
#else
    ret = read_symbols(fp, exe, 0);
/*    reflection_dlopen(fp, exe, "main");*/
#endif

    return ret;
}

#if defined(WIN32)
static int dlopen_arch(regsym_cb fp, const char* file, const char* func) {
    HMODULE h;
    int i = 0;
    char buf[256];

    if (!strchr(file, '.')) {
        strcpy(buf, file);
        strcat(buf, ".dll");
        file = buf;
    }

    h = LoadLibrary(file);
    if (!h) {
        winerror("GetModuleHandle");
        return 1;
    }
    return init_arch(fp);
}
#else
static int dlopen_arch(regsym_cb fp, const char* file, const char* func) {
    void* dlhandle;
    char buf[256];

    if (!strchr(file, '/')) {
        if (strncmp(file, "lib", 3) != 0) {
            strcpy(buf, "lib");
            strcat(buf, file);
            file = buf;
        }
    }
    if (!strchr(file, '.')) {
        if (buf != file) strcpy(buf, file);
#ifdef __APPLE__
        strcat(buf, ".dylib");
#else
        strcat(buf, ".so");
#endif
        file = buf;
    }

    if (!(dlhandle = dlopen(file, RTLD_NOW))) {
        perror("cannot dlopen");
        return 1;
    }

#ifdef HAVE_DLINFO
    {
        struct link_map* lmap;

        if (dlinfo(dlhandle, RTLD_DI_LINKMAP, &lmap) == 0) {
            while (lmap) {
                if (read_symbols(fp, lmap->l_name, lmap->l_addr)) return 1;
                lmap = lmap->l_next;
            }
            return 0;
        }
        else {
            perror("dlinfo");
            return 1;
        }
    }
#else
    {
        void* fp;
        Dl_info info;
        if (!func) {
            perror("2nd argument is needed in this environment");
            return 1;
        }
        if (!(fp = dlsym(dlhandle, func))) {
            perror("dlsym");
            return 1;
        }
        if (!dladdr(fp, &info)) {
            perror("dladdr");
            return 1;
        }
        read_symbols(fp, info.dli_fname, (int)info.dli_fbase);

        return 0;
    }
#endif
}
#endif

int reflection_dlopen(regsym_cb fp, const char* file, const char* func) {
    return dlopen_arch(fp, file, func);
}
